@@ -2,11 +2,11 @@
  *  html.c
  *  mod_musicindex
  *
- *  $Id: html.c 674 2006-09-03 14:51:46Z varenet $
+ *  $Id: html.c 755 2007-04-29 20:47:50Z varenet $
  *
  *  Created by Thibaut VARENE on Thu Mar 20 2003.
  *  Copyright (c) 2003-2006 Regis BOUDIN
- *  Copyright (c) 2003-2006 Thibaut VARENE
+ *  Copyright (c) 2003-2007 Thibaut VARENE
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License version 2.1,
@@ -20,8 +20,8 @@
  *
  * @author Regis Boudin
  * @author Thibaut Varene
- * @version $Revision: 674 $
- * @date 2003-2006
+ * @version $Revision: 755 $
+ * @date 2003-2007
  *
  * That file groups all functions related to HTML and assimilated
  * generation, that is to say functions that are used to
@@ -40,7 +40,11 @@
 
 #include <http_core.h>
 
+#ifdef TM_IN_SYS_TIME
+#include <sys/time.h>
+#else
 #include <time.h> /* For podcast informations */
+#endif
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -53,12 +57,6 @@
 
 static const char const Gfavicon[] = "sound.png";
 static const char const Gcd_icon[] = "general.png";
-static const char const Gfetch_icon[] = "fetch.png";
-static const char const Gsound_icon[] = "sound.png";
-static const char const Gshuffle_icon[] = "shuffle.png";
-static const char const Grss_icon [] = "rss.png";
-static const char const Garrow[] = "right_arrow.gif";
-static const char const Gsmall_cd_icon[] = "directory.png";
 static const char const Gorigcss[] = "musicindex.css";
 /** Cover icon names to look for in a directory */
 static const char *const Gcovericns[] = {
@@ -75,7 +73,7 @@ static const char *const Gcovericns[] = {
  * Formats the list of songs to be sent.
  *
  * This function formats a list of tracks (if any) to be sent
- * to the client, as found in the p list passed in argument.
+ * to the client, as found in the @a p list passed in argument.
  * Depending on config options, it allows streaming/shuffle etc.
  *
  * @todo Sort option for recursive search?
@@ -88,11 +86,14 @@ static const char *const Gcovericns[] = {
  *
  * @return The total number of fields in a row.
  */
-static unsigned short list_songs(request_rec *r, const mu_ent *const p, const mu_config *const conf, const int customlist)
+static unsigned short
+list_songs(request_rec *r, const mu_ent *const p, const mu_config *const conf,
+		const int customlist)
 {
 	const mu_ent *q = p;
 	register unsigned short fld = 0, j = 0;
 	const char *current = NULL, *new = NULL;
+	char *codeduri = NULL;
 
 	ap_rputs("  <tr class=\"title\">\n", r);
 
@@ -155,6 +156,10 @@ static unsigned short list_songs(request_rec *r, const mu_ent *const p, const mu
 				current = "Size";
 				new = _("Size");
 				break;
+			case SB_FILENAME:
+				current = "Filename";
+				new = _("Filename");
+				break;
 			default:
 				continue;
 		}
@@ -171,7 +176,7 @@ static unsigned short list_songs(request_rec *r, const mu_ent *const p, const mu
 		/* XXX sorting is disabled for search results */
 		ap_rvputs(r, "   <th class=\"", current, "\">", NULL);
 		if ((customlist == 0) && !(conf->search))
-			ap_rprintf(r, "<a href=\"?sort=%c\">%s</a>", conf->fields[fld], new);
+			ap_rprintf(r, "<a href=\"?sort=%d\">%s</a>", conf->fields[fld], new);
 		else
 			ap_rputs(new, r);
 		ap_rputs("</th>\n", r);
@@ -184,22 +189,25 @@ static unsigned short list_songs(request_rec *r, const mu_ent *const p, const mu
 	current = "";
 	new = current;
 
-	for (q=p; q!= NULL; q=q->next) {	/* XXX doc */
+	for (q=p; q!= NULL; q=q->next) {
 		if (q->filetype < 0)
 			continue;
 
+		/* If we have a recursive search request we have some work to do */
 		if ((conf->search != NULL) && (customlist == 0) &&
-			(conf->options & MI_RECURSIVE) == MI_RECURSIVE) {
+			(conf->options & MI_RECURSIVE)) {
 			new = ap_make_dirstr_parent(r->pool, q->file);
 
+			/* We just changed directories (during search, entries
+			 * are sorted on a per-dir basis) so print a nice header */
 			if (strcmp(current, new)) {
 				current = new;
-				ap_rputs("  <tr class=\"title\">\n", r);
-				ap_rprintf(r, "   <th align=\"left\" colspan=\"%d\">", fld+1);	/* "Select" => +1 */
+				ap_rprintf(r, "  <tr class=\"title\">\n"
+					"   <th align=\"left\" colspan=\"%d\">", fld+1);	/* "Select" => +1 */
 				if ((current[0] == '\0') || (current[1] == '\0'))	/* current dir is either "" or "/" */
 					ap_rputs(_("In Current Directory"), r);
 				else
-					ap_rvputs(r, _("In "), "<a href=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, current)), "\">", ap_escape_html(r->pool, current), "</a>", NULL);
+					ap_rvputs(r, _("In"), " <a href=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, current)), "\">", ap_escape_html(r->pool, current), "</a>", NULL);
 
 				ap_rputs("</th>\n"
 					"  </tr>\n", r);
@@ -215,28 +223,31 @@ static unsigned short list_songs(request_rec *r, const mu_ent *const p, const mu
 		/* prepare the "Select" panel, if any */
 		if ((customlist != 0) || (conf->options & (MI_ALLOWDWNLD | MI_ALLOWSTREAM)) ||
 			((conf->search) && (conf->options & MI_RECURSIVE))) {
+			/* this should save some memory */
+			codeduri = (char *)realloc(codeduri, 1 + ap_base64encode_len(strlen(q->uri)));
+
 			ap_rputs("   <td class=\"Select\">\n", r);
 
-			if (q->flags & EF_ALLOWSTREAM) {	/* Display checkbox */
-				const char *locator = q->uri;	/* XXX simplify */
-				char *codeduri = (char *)ap_palloc(r->pool, 1 + ap_base64encode_len(strlen(locator)));
-				if (codeduri)
-					ap_base64encode(codeduri, locator, strlen(locator));
+			/* Display checkbox */
+			if (codeduri) {
+				ap_base64encode(codeduri, q->uri, strlen(q->uri));
 				/* we base64 encode the uri to avoid nasty special chars problems */
 				ap_rvputs(r, "    <input type=\"checkbox\" name=\"file\" value=\"",
 					codeduri, "\" />\n", NULL);
 			}
 
-			if (customlist == 0) {
+			if (!customlist) {
 				if (q->flags & EF_ALLOWDWNLD)	/* Display [download] */
-					ap_rvputs(r, "    <a href=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)), "\">"
-						"<img alt=\"[D]\" title=\"", _("Download"), "\" src=\"", conf->directory, "/",
-						Gfetch_icon, "\" /></a>\n", NULL);
+					ap_rvputs(r, "    <a class=\"dld\" href=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)), "\" title=\"", _("Download"), "\">&nbsp;</a>\n", NULL);
 
 				if (q->flags & EF_ALLOWSTREAM)	/* Display [stream] */
-					ap_rvputs(r, "    <a href=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)), "?stream\">"
-						"<img alt=\"[S]\" title=\"", _("Stream"), "\" src=\"", conf->directory, "/",
-						Gsound_icon, "\" /></a>\n", NULL);
+					ap_rvputs(r, "    <a class=\"stream\" href=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)), "?stream\" title=\"", _("Stream"), "\">&nbsp;</a>\n", NULL);
+			} else {	/* in a custom list */
+				if (q->flags & EF_ALLOWTARBALL)	/* Tell that tarball download will work with this file */
+					ap_rvputs(r, "    <span class=\"dldok\" title=\"", _("Download Allowed"), "\">&nbsp;</span>\n", NULL);
+
+				if (q->flags & EF_ALLOWSTREAM)	/* Tell that streaming will work with this file */
+					ap_rvputs(r, "    <span class=\"streamok\" title=\"", _("Stream Allowed"), "\">&nbsp;</span>\n", NULL);
 			}
 
 			ap_rputs("   </td>\n", r);
@@ -283,7 +294,7 @@ static unsigned short list_songs(request_rec *r, const mu_ent *const p, const mu
 					if (conf->options & MI_QUICKPL)
 						break;
 					if (q->freq)
-						ap_rprintf(r, "   <td class=\"Freq\"><acronym title=\"kHz\">%.1f</acronym></td>\n", (float)q->freq / 1000);
+						ap_rprintf(r, "   <td class=\"Freq\"><acronym title=\"kHz\">%d.%d</acronym></td>\n", q->freq / 1000, q->freq % 1000 / 100);
 					else
 						ap_rputs( "   <td></td>\n", r);
 					break;
@@ -306,7 +317,10 @@ static unsigned short list_songs(request_rec *r, const mu_ent *const p, const mu
 					ap_rprintf(r, "   <td class=\"Genre\">%s</td>\n", q->genre ? ap_escape_html(r->pool, q->genre) : "");
 					break;
 				case SB_SIZE:
-					ap_rprintf(r, "   <td class=\"Size\"><acronym title=\"MB\">%.2f</acronym></td>\n", ((float)(q->size>>10))/1024);
+					ap_rprintf(r, "   <td class=\"Size\"><acronym title=\"MB\">%ld.%ld</acronym></td>\n", q->size>>20, (q->size>>10)%1024/100);
+					break;
+				case SB_FILENAME:
+					ap_rvputs(r, "   <td class=\"Filename\">", ap_escape_html(r->pool, q->filename), "</td>\n", NULL);
 					break;
 				default:
 					break;
@@ -315,6 +329,7 @@ static unsigned short list_songs(request_rec *r, const mu_ent *const p, const mu
 		/* at the end of the loop, 'fld' contains the number of fields */
 		ap_rputs("  </tr>\n", r);
 	}
+	free(codeduri);
 	return fld;
 }
 
@@ -324,18 +339,23 @@ static unsigned short list_songs(request_rec *r, const mu_ent *const p, const mu
  * This function does all the URL formatting, in the form:
  * http://[user:][passwd\@]hostname[:port]/uri/
  *
+ * @bug If the admin sets a conf->iceserver too big (>256) we're gonna be in
+ *	trouble. Same thing goes if hostname+auth+etc is too long...
+ *
  * @param r Apache request_rec struct to handle connection details.
  * @param uri A string (basically a path) to be sent to form a complete url.
  * @param command An eventual command (like ?action=playall).
  * @param conf MusicIndex configuration paramaters struct.
+ * @param html 1 ? html output : normally escaped URL output.
  */
 static void send_url(request_rec *r, const char *const uri, const char *const command,
-			const mu_config *const conf, short html)
+			const mu_config *const conf, const short html)
 {
-	char prefix[MAX_PREFIX];
-	char str_port[6];	/* 65536 + '\0' */
+	static char prefix[MAX_PREFIX];
+	static char str_port[7], strtmp[6];	/* (:)65536 + '\0' */
 	char *bufcoded, *decoded;
 	unsigned short l;
+	unsigned register i=0, j=6;
 
 	/* If we have an icecast server, it takes precedence, in a 'staticdir' fashion.
 	If we're sending an RSS feed, it's up to us to send the ?stream links. */
@@ -373,8 +393,17 @@ static void send_url(request_rec *r, const char *const uri, const char *const co
 		 * checking whether it's really needed or not is just too
 		 * much overhead.
 		 * ap_get_server_port() should hopefully deal with UseCanonicalName
-		 * for us, and thus redirected ports will be properly handled. */
-		sprintf(str_port, ":%u", ap_get_server_port(r));
+		 * for us, and thus redirected ports will be properly handled.
+		 * Custom 'snprintf' implementation for speed & fun */
+		l = ap_get_server_port(r);
+		while (--j && l) {
+			strtmp[i++] = "0123456789"[l%10];
+			l /= 10;
+		}
+		str_port[0] = ':';
+		for (j=1; i; j++)
+			str_port[j] = strtmp[--i];
+		str_port[j] = '\0';
 		strcat(prefix, str_port);
 	}
 
@@ -395,18 +424,16 @@ static void send_url(request_rec *r, const char *const uri, const char *const co
  *
  * @param r Apache request_rec struct to handle connection details.
  * @param conf MusicIndex configuration paramaters struct.
- * 
- * @bug When we'll allow conf->directory to be modified, have to escape html/uri.
+ *
+ * @bug When we'll allow #mu_config->directory to be modified, have to escape html/uri.
  */
 void send_head(request_rec *r, const mu_config *const conf)
 {
-	short i;
+	register short i;
 	const char *dir;
 	char *uri, *u;
 	request_rec *subreq = NULL;
-#ifdef NEW_HEADER /* New ! :) */
 	const mu_config *localconf = NULL;
-#endif
 
 	ap_rputs("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
 		"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" "
@@ -428,22 +455,26 @@ void send_head(request_rec *r, const mu_config *const conf)
 				if ((dstruct->d_name[0] != '.') &&
 					(ext = strrchr(dstruct->d_name, '.')) &&
 					(!strncmp(ext+1, "css", 3))) {
-
+					/* XXX we don't make sure the files are R_OK */
 					if (!strcmp(dstruct->d_name, Gorigcss))
 						continue;	/* this will be dealt with later */
 					if (!strcmp(dstruct->d_name, conf->css))
 						ap_rputs(" <link rel=\"stylesheet\" title=\"default\"", r);
 					else
-						ap_rvputs(r, " <link rel=\"alternate stylesheet\" title=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, dstruct->d_name)), "\"", NULL);
+						ap_rvputs(r, " <link rel=\"alternate stylesheet\" title=\"",
+							ap_escape_html(r->pool, ap_escape_uri(r->pool, dstruct->d_name)),
+							"\"", NULL);
 
-					ap_rvputs(r, " type=\"text/css\" href=\"", conf->directory, "/", ap_escape_html(r->pool, ap_escape_uri(r->pool, dstruct->d_name)), "\" />\n", NULL);
+					ap_rvputs(r, " type=\"text/css\" href=\"", conf->directory, "/",
+						ap_escape_html(r->pool, ap_escape_uri(r->pool, dstruct->d_name)),
+						"\" />\n", NULL);
 				}
 			}
 			closedir(dir);
 		}
 		ap_destroy_sub_req(subreq);
 	}
-		
+
 	/* the original CSS */
 	if (!strcmp(Gorigcss, conf->css))
 		ap_rputs(" <link rel=\"stylesheet\" title=\"default\"", r);
@@ -484,7 +515,7 @@ void send_head(request_rec *r, const mu_config *const conf)
 			break;
 		}
 	}
-	
+
 	if (i > 0)
 		ap_rvputs(r, conf->directory, "/", Gcd_icon, NULL);
 
@@ -497,7 +528,7 @@ void send_head(request_rec *r, const mu_config *const conf)
 
 	uri = ap_pstrdup(r->pool, r->uri);
 
-#ifdef NEW_HEADER /* New ! :) */	/* XXX Could use some doc... :P */
+	/* XXX Could use some doc... :P */
 	for (u = uri; *u != '\0'; u++) {
 		dir = u;
 		while ((*u != '/') && (*u != '\0'))
@@ -524,73 +555,44 @@ void send_head(request_rec *r, const mu_config *const conf)
 			*u = '/';
 
 			if (*(u+1) != '\0')
-				ap_rvputs(r, "   <img src=\"", conf->directory, "/",
-					Garrow, "\" alt=\"=>\" />\n", NULL);
+				ap_rputs("   <span class=\"rarrow\">&nbsp;</span>\n", r);
 		}
 
 		if (subreq != NULL)
 			ap_destroy_sub_req(subreq);
 	}
-#else /* Here to have a fallback solution if there is a problem with the new code */
-	/* XXX Maybe we can trash this? */
-	for (u = uri; *u != '\0'; u++) {
-		dir = u;
-		while ((*u != '/') && (*u != '\0'))
-			u++;
-
-		if (u == uri)
-			dir = conf->title;
-		else
-			ap_rvputs(r, "     <img src=\"", conf->directory, "/",
-				Garrow, "\" alt=\"=>\" />\n", NULL);
-
-		*u = '\0';
-		ap_rvputs(r, "   <a href=\"", uri, "/\">", dir, "</a>\n", NULL);
-		*u = '/';
-	}
-#endif
 
 	ap_rputs("  </h1>\n", r);
 
 	if (conf->options & MI_ALLOWSTREAM) {
 		ap_rvputs(r, "  <a class=\"shuffle\" "
-			"href=\"?option=recursive&amp;option=shuffle&amp;action=playall\">",
-			"<img alt=\"[R]\" title=\"", _("Shuffle All"), "\" src=\"",
-			conf->directory, "/", Gshuffle_icon,
-			"\" /></a>\n"
+			"href=\"?option=recursive&amp;option=shuffle&amp;action=playall\""
+			" title=\"", _("Shuffle All"), "\">&nbsp;</a>\n"
 			"  <a class=\"stream\" "
-			"href=\"?option=recursive&amp;action=playall\">",
-			"<img alt=\"[S]\" title=\"", _("Stream All"), "\" src=\"",
-			conf->directory, "/", Gsound_icon,
-			"\" /></a>\n", NULL);
+			"href=\"?option=recursive&amp;action=playall\""
+			" title=\"", _("Stream All"), "\">&nbsp;</a>\n", NULL);
 	}
 
 	if (conf->options & MI_ALLOWTARBALL) {
 		ap_rvputs(r, "  <a class=\"tarball\" "
-			"href=\"?option=recursive&amp;action=tarball\">",
-			"<img alt=\"[D]\" title=\"", _("Download All"), "\" src=\"",
-			conf->directory, "/", Gfetch_icon,
-			"\" /></a>\n", NULL);
+			"href=\"?option=recursive&amp;action=tarball\""
+			" title=\"", _("Download All"), "\">&nbsp;</a>\n", NULL);
 	}
 
 	if (conf->rss_items > 0) {
 		ap_rvputs(r, "    <a class=\"rss\" "
-			"href=\"?action=RSS\">",
-			"<img alt=\"[RSS]\" title=\"", _("RSS"), "\" src=\"",
-			conf->directory, "/", Grss_icon,
-			"\" /></a>\n", NULL);
+			"href=\"?action=RSS\""
+			" title=\"", _("RSS"), "\">&nbsp;</a>\n", NULL);
 	}
 
 	/* XXX to integrate in a better way */
-	ap_rvputs(r, "    <br /><a class=\"rss\" "
+	ap_rvputs(r, "    <br /><a class=\"rdir\" "
 			"href=\"?action=randomdir\">[",
-			_("Random subdirectory..."), "]</a>\n", NULL);
-
-	ap_rputs(" </div>\n", r);
+			_("Random subdirectory..."), "]</a>\n </div>\n", NULL);
 
 	/* displays a search box if option is activated */
 	/* XXX collision avec le header, a regler */
- 	if (conf->options & MI_ALLOWSEARCH) {
+	if (conf->options & MI_ALLOWSEARCH) {
 		ap_rvputs(r,
 			" <form method=\"post\" action=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, r->uri)), "\""
 				" enctype=\"application/x-www-form-urlencoded\""
@@ -598,8 +600,8 @@ void send_head(request_rec *r, const mu_config *const conf)
 			"  <p>\n"
 			"   <input type=\"text\" name=\"search\" />\n"
 			"   <br />\n"
-			"   <input type=\"submit\" name=\"action\" value=\"Search\" />\n"
-			"   <input type=\"submit\" name=\"action\" value=\"Recursive Search\" />\n"	/* XXX youpi on peut pas i18n-iser... */
+			"   <button type=\"submit\" name=\"action\" value=\"Search\">", _("Search"), "</button>\n"
+			"   <button type=\"submit\" name=\"action\" value=\"RecursiveSearch\">", _("Recursive Search"), "</button>\n"
 			"   <input type=\"hidden\" name=\"action\" value=\"Search\" />\n"
 			"  </p>\n"
 			" </form>\n",
@@ -619,20 +621,21 @@ void send_head(request_rec *r, const mu_config *const conf)
  * list (if any) of the available directories in the current folder.
  * Depending on config options, it allows streaming/shuffle etc.
  *
+ * @remark I tried moving to purely semantic coding (using either div or li,
+ *	none worked. It seems CSS/browsers can't render properly a list of
+ *	element of fixed width but variable height in a variable "columns"
+ *	context: if an element has a greater height than the one immediately
+ *	after on the same "row", then the next row will start at the right of
+ *	this longer element instead of starting back at the left of the page.
+ *
  * @param r Apache request_rec struct to handle connection details.
- * @param p struct mu_ent.
+ * @param pack A pack of music entries.
  * @param conf MusicIndex configuration paramaters struct.
  */
-void send_directories(request_rec *r, const mu_ent *const p, const mu_config *const conf)
+void send_directories(request_rec *r, const mu_pack *const pack, const mu_config *const conf)
 {
 	const mu_ent *q;
-	unsigned short dircnt = 0, nb = 0;
-	unsigned register short i;
-	char temp[MAX_STRING];
-
-	/* We count the number of directories for later use */
-	for (q = p; q && (q->filetype < 0); q = q->next)
-		nb++;
+	unsigned short dircnt = 0, nb = pack->dirnb;
 
 	if (nb == 0)
 		return;
@@ -641,88 +644,48 @@ void send_directories(request_rec *r, const mu_ent *const p, const mu_config *co
 	ap_rprintf(r, _("Music Directories (%d)"), nb);
 	ap_rputs("</h2>\n\n<table id=\"directories\">\n", r);
 
-	for (q=p; q && (q->filetype < 0); q = q->next) {
-		char *tend = ap_cpystrn(temp, q->file, MAX_STRING-1);
-		/* Copies at most numbytes of str to buf. Differs from strncpy()
-		in that buf is always null terminated, but is not null filled.
-		Therefore, buf should always be at least numbytes + 1 bytes long.
-		Returns a pointer to the terminating '\0'. */
-
-		/* remove trailing "/" */
-		tend--;
-		if (*tend == '/')
-			*tend = '\0';
-
-		/* remove underscores from directories display to help formatting (wrapping) */
-		for (i=0; temp[i]; i++)
-			if (temp[i] == '_')
-				temp[i] = ' ';
-
+	for (q=pack->head; q && (q->filetype < 0); q = q->next) {
 		/* start a new directories row */
 		if (dircnt++ == 0)
 			ap_rputs(" <tr>\n", r);
 
 		ap_rvputs(r, "  <td>\n"
-			"   <a href=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)), NULL);
-
-		ap_rputs("\"><img alt=\"\" src=\"", r);
-#ifdef SHOW_THUMBNAILS
-/* XXX tv: This is just wrong. It doesn't rescale the picture, hence, loading 20 dirs with a 200kB pic each is AWFUL
-Not mentionning the fact that letting the browser do the scaling leads often to ugly and unreadable results.
-That's IMHO a bad feature */
-		/* TODO : optimize to use less memory */
-		if (access(ap_pstrcat(r->pool, r->filename, "/", q->file, "cover.png", NULL), R_OK) == 0)
-			ap_rvputs(r, q->file, "cover.png", NULL);
-		else if (access(ap_pstrcat(r->pool, r->filename, "/", q->file, "cover.jpg", NULL), R_OK) == 0)
-			ap_rvputs(r, q->file, "cover.jpg", NULL);
-		else if (access(ap_pstrcat(r->pool, r->filename, "/", q->file, "cover.gif", NULL), R_OK) == 0)
-			ap_rvputs(r, q->file, "cover.gif", NULL);
-		else
-#endif
-			ap_rvputs(r, conf->directory, "/", Gsmall_cd_icon, NULL);
-
-		ap_rputs("\" /></a>\n", r);
-
-		ap_rvputs(r, "   <div>\n"
+			"   <a href=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)),
+			"\" class=\"bigdir\" title=\"", _("Open"), "\">&nbsp;</a>\n"
+			"   <div>\n"
 			"    <a href=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)), "\">",
-	    		temp, "</a><br />\n", NULL);
+			ap_escape_html(r->pool, q->title), "</a>", NULL);
 
-		/* show various useful links when needed */
-		if (q->flags & EF_ALLOWSTREAM) {
-			ap_rvputs(r, "    <a class=\"shuffle\" href=\"",
-				ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)),
-				"?option=recursive&amp;option=shuffle&amp;action=playall\">",
-				"<img alt=\"[R]\" title=\"", _("Shuffle"), "\" src=\"",
-				conf->directory, "/", Gshuffle_icon,
-				"\" /></a>\n", NULL);
-
-			ap_rvputs(r, "    <a class=\"stream\" href=\"",
-				ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)),
-				"?option=recursive&amp;action=playall\">",
-				"<img alt=\"[S]\" title=\"", _("Stream"), "\" src=\"",
-				conf->directory, "/", Gsound_icon,
-				"\" /></a>\n", NULL);
+		if (PRETTY_FOLDERS) {
+			/* show various useful links when needed */
+			ap_rputs("<br />\n", r);
+			if (q->flags & EF_ALLOWSTREAM) {
+				ap_rvputs(r, "    <a class=\"shuffle\" href=\"",
+					ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)),
+					"?option=recursive&amp;option=shuffle&amp;action=playall\""
+					" title=\"", _("Shuffle"), "\">&nbsp;</a>\n"
+					"    <a class=\"stream\" href=\"",
+					ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)),
+					"?option=recursive&amp;action=playall\""
+					" title=\"", _("Stream"), "\">&nbsp;</a>\n", NULL);
+			}
+	
+			if (q->flags & EF_ALLOWTARBALL) {
+				ap_rvputs(r, "    <a class=\"tarball\" href=\"",
+					ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)),
+					"?option=recursive&amp;action=tarball\""
+					" title=\"", _("Download"), "\">&nbsp;</a>\n", NULL);
+			}
+	
+			if (q->flags & EF_ALLOWRSS) {
+				ap_rvputs(r, "    <a class=\"rss\" href=\"",
+					ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)),
+					"?action=RSS\""
+					" title=\"", _("RSS"), "\">&nbsp;</a>\n", NULL);
+			}
 		}
 
-		if (q->flags & EF_ALLOWTARBALL) {
-			ap_rvputs(r, "    <a class=\"tarball\" href=\"",
-				ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)),
-				"?option=recursive&amp;action=tarball\">",
-				"<img alt=\"[D]\" title=\"", _("Download"), "\" src=\"",
-				conf->directory, "/", Gfetch_icon,
-				"\" /></a>\n", NULL);
-		}
-
-		if (q->flags & EF_ALLOWRSS) {
-			ap_rvputs(r, "    <a class=\"rss\" href=\"",
-				ap_escape_html(r->pool, ap_escape_uri(r->pool, q->file)),
-				"?action=RSS\">",
-				"<img alt=\"[RSS]\" title=\"", _("RSS"), "\" src=\"",
-				conf->directory, "/", Grss_icon,
-				"\" /></a>\n", NULL);
-		}
-
-		ap_rputs("   </div>\n"
+		ap_rputs("\n   </div>\n"
 			"  </td>\n",r);
 
 		/* end the directories row */
@@ -743,47 +706,38 @@ That's IMHO a bad feature */
  * Sends track listing for the current folder.
  *
  * This function takes care of preparing and sending to the client the
- * list (if any) of the available songs in the current p list.
+ * list (if any) of the available songs in the current @a pack list.
  * Depending on config options, it allows streaming/shuffle etc.
  *
  * @param r Apache request_rec struct to handle connection details.
- * @param p struct mu_ent.
+ * @param pack A pack of music entries.
  * @param conf MusicIndex configuration paramaters struct.
  */
-void send_tracks(request_rec *r, const mu_ent *const p, const mu_config *const conf)
+void send_tracks(request_rec *r, const mu_pack *const pack, const mu_config *const conf)
 {
-	const mu_ent *q;
+	const mu_ent *q = pack->head;
 	register unsigned short nb = 0;
-	unsigned short fld;
+	unsigned long fnb = pack->filenb;
+	const unsigned char *const restrict order = conf->order;
 
-	/* We count the number of songs for later use */
-	for (q = p; q; q = q->next) {
-		if (q->filetype >= 0)	/* XXX try to get rid of this test: on peut avoir des 'dirs' ou des fichiers non reconnus dans cette liste de mu_ents? */
-			nb++;
-	}
-
-	if (nb == 0)
+	if (fnb == 0)
 		return;
 
 	ap_rputs("<!-- begin tracks -->\n<h2>", r);
 
 	if (conf->search)
-		ap_rprintf(r, _("Result List (%d)"), nb);
+		ap_rprintf(r, _("Result List (%ld)"), fnb);
 	else
-		ap_rprintf(r, _("Song List (%d)"), nb);
+		ap_rprintf(r, _("Song List (%ld)"), fnb);
 
-	ap_rputs("</h2>\n\n", r);
+	ap_rvputs(r, "</h2>\n\n<form method=\"post\" action=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, r->uri)), "\" "
+		"enctype=\"application/x-www-form-urlencoded\" id=\"tracks\">\n <table>\n", NULL);
 
-	ap_rvputs(r, "<form method=\"post\" action=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, r->uri)), "\" "
-		"enctype=\"application/x-www-form-urlencoded\" id=\"tracks\">\n", NULL);
-
-	ap_rputs(" <table>\n", r);
-
-	fld = list_songs(r, p, conf, 0);
+	nb = list_songs(r, q, conf, 0);
 
 	/* Do NOT show 'select all' if there's nothing to select... */
 	if ((conf->search) && (conf->options & MI_ALLOWSTREAM)) {
-		ap_rprintf(r, "  <tr class=\"title\"><th align=\"left\" colspan=\"%d\">\n", fld+1); /* Select => +1 */
+		ap_rprintf(r, "  <tr class=\"title\"><th align=\"left\" colspan=\"%d\">\n", nb+1); /* Select => +1 */
 		ap_rputs("   <input type=\"checkbox\" name=\"all\" onClick=\"for(var i=0;i<this.form.elements.length;i++){var inpt=this.form.elements[i];"
 			"var m=inpt.name.match(/-/g);if((inpt.name.substr(0,4)=='file') && (m<1)) inpt.checked=this.form.all.checked}\" />\n"
 			"Select All</th>\n"
@@ -792,27 +746,37 @@ void send_tracks(request_rec *r, const mu_ent *const p, const mu_config *const c
 
 	ap_rputs(" </table>\n", r);
 
-	/* Do NOT show buttons if we do not allow anything but listing or downloading! */
+	/* Do NOT show buttons if we do not allow anything but listing or individual downloading! */
 	if ((conf->options & (MI_ALLOWSTREAM|MI_ALLOWTARBALL))) {
-		ap_rvputs(r, " <div>\n"
-			"  <input type=\"hidden\" name=\"sort\" value=\"", conf->order,"\" />\n"
-			"  <input type=\"submit\" name=\"action\" value=\"Add To Playlist\" class=\"playlist\" />\n", NULL); /* XXX i18n DTC */
+		ap_rputs(" <div>\n"
+			"  <input type=\"hidden\" name=\"sort\" value=\"", r);
+		for(nb=0; order[nb] && (nb<SB_MAX); nb++)
+			ap_rprintf(r, "%c", order[nb]+'`');	/* 1+'`' = 'a'. shift everything to ascii small letters */
+		ap_rvputs(r, "\" />\n"
+			"  <button type=\"submit\" name=\"action\" value=\"AddToPlaylist\" class=\"playlist\">",
+			_("Add To Playlist"), "</button>\n", NULL);
 
 		if (conf->search == NULL) {
-			ap_rputs("  <input type=\"submit\" name=\"action\" value=\"Add All To Playlist\" class=\"playlist\" />\n", r);
+			ap_rvputs(r, "  <button type=\"submit\" name=\"action\" value=\"AddAllToPlaylist\" class=\"playlist\">",
+				_("Add All To Playlist"), "</button>\n", NULL);
 			if (conf->options & MI_ALLOWSTREAM)
-				ap_rputs("  <input type=\"submit\" name=\"action\" value=\"Shuffle All\" />\n"
-				"  <input type=\"submit\" name=\"action\" value=\"Play All\" />\n", r);	/* XXX idem */
+				ap_rvputs(r, "  <button type=\"submit\" name=\"action\" value=\"ShuffleAll\">",
+				_("Shuffle All"), "</button>\n"
+				"  <button type=\"submit\" name=\"action\" value=\"PlayAll\">",
+				_("Play All"), "</button>\n", NULL);
 		}
 
 		if (conf->options & MI_ALLOWSTREAM)
-			ap_rputs("  <input type=\"submit\" name=\"action\" value=\"Play Selected\" />\n", r);
+			ap_rvputs(r, "  <button type=\"submit\" name=\"action\" value=\"PlaySelected\">",
+				_("Play Selected"), "</button>\n", NULL);
 
 		if (conf->options & MI_ALLOWTARBALL)
-			ap_rputs("  <input type=\"submit\" name=\"action\" value=\"Download Selected\" />\n"
-				"  <input type=\"submit\" name=\"action\" value=\"Download All\" />\n", r);
+			ap_rvputs(r, "  <button type=\"submit\" name=\"action\" value=\"DownloadSelected\">",
+				_("Download Selected"), "</button>\n"
+				"  <button type=\"submit\" name=\"action\" value=\"DownloadAll\">",
+				_("Download All"), "</button>\n", NULL);
 
-		ap_rputs(" </div>\n", r);	/* XXX et encore */
+		ap_rputs(" </div>\n", r);
 	}
 
 	ap_rputs("</form>\n"
@@ -828,45 +792,42 @@ void send_tracks(request_rec *r, const mu_ent *const p, const mu_config *const c
  * retrieved from a cookie.
  * Depending on config options, it allows streaming/shuffle etc.
  *
+ * @warning Even though the buttons are here (Download All, Stream All...),
+ *	custom playlists will always respect individual files permissions,
+ *	meaning that one can end up with an empty playlist or tarball even
+ *	though the list itself isn't empty.
+ *
  * @param r Apache request_rec struct to handle connection details.
- * @param p struct mu_ent.
+ * @param pack A pack of music entries.
  * @param conf MusicIndex configuration paramaters struct.
  */
-void send_customlist(request_rec *r, const mu_ent *const p, const mu_config *const conf)
+void send_customlist(request_rec *r, const mu_pack *const pack, const mu_config *const conf)
 {
-	const mu_ent *q;
-	register unsigned short nb = 0;
+	const mu_ent *q = pack->head;
+	register unsigned short nb = pack->filenb;
 
-	if (p == NULL)
+	if (!q)
 		return;
 
-	/* We count the number of songs for later use */
-	for (q = p; q; q = q->next)
-		nb++;
-
 	ap_rputs("<!-- begin custom -->\n"
 		"<h2>", r);
 	ap_rprintf(r, _("Custom Playlist (%d)"), nb);
 	ap_rputs("</h2>\n\n", r);
 
 	ap_rvputs(r, " <form method=\"post\" action=\"", ap_escape_html(r->pool, ap_escape_uri(r->pool, r->uri)), "\" "
-		"enctype=\"application/x-www-form-urlencoded\" id=\"custom\">\n", NULL);
+		"enctype=\"application/x-www-form-urlencoded\" id=\"custom\">\n  <table>\n", NULL);
 
-	ap_rputs("  <table>\n", r);
+	list_songs(r, q, conf, 1);
 
-	list_songs(r, p, conf, 1);
-
-	ap_rputs("  </table>\n"
-		"  <div>\n", r);
-
-	ap_rputs("   <input type=\"submit\" name=\"action\" value=\"Remove from Playlist\" class=\"playlist\" />\n"
-		"   <input type=\"submit\" name=\"action\" value=\"Clear Playlist\" class=\"playlist\" />\n", r);
-
-	if ((conf->options & MI_ALLOWSTREAM) != 0)
-	ap_rputs("   <input type=\"submit\" name=\"action\" value=\"Stream Playlist\" class=\"playlist\" />\n", r);
-
-	if ((conf->options & MI_ALLOWTARBALL) != 0)
-	ap_rputs("   <input type=\"submit\" name=\"action\" value=\"Download Playlist\" class=\"playlist\" />\n", r);
+	ap_rvputs(r, "  </table>\n"
+		"  <div>\n   <button type=\"submit\" name=\"action\" value=\"RemoveFromPlaylist\" class=\"playlist\">",
+		_("Remove From Playlist"), "</button>\n"
+		"   <button type=\"submit\" name=\"action\" value=\"ClearPlaylist\" class=\"playlist\">",
+		_("Clear Playlist"), "</button>\n"
+		"   <button type=\"submit\" name=\"action\" value=\"StreamPlaylist\" class=\"playlist\">",
+		_("Stream Playlist"), "</button>\n"
+		"   <button type=\"submit\" name=\"action\" value=\"DownloadPlaylist\" class=\"playlist\">",
+		_("Download Playlist"), "</button>\n", NULL);
 
 	ap_rputs("  </div>\n"
 		" </form>\n"
@@ -881,19 +842,23 @@ void send_customlist(request_rec *r, const mu_ent *const p, const mu_config *con
  * available: length, artist, title, album) and sends it to the client.
  *
  * @param r Apache request_rec struct to handle connection details.
- * @param p struct mu_ent.
+ * @param pack A pack of music entries.
  * @param conf MusicIndex configuration paramaters struct.
  */
-void send_playlist(request_rec *r, const mu_ent *const p, const mu_config *const conf)
+void send_playlist(request_rec *r, const mu_pack *const pack, const mu_config *const conf)
 {
-	const mu_ent *q = p;
+	const mu_ent *q = pack->head;
 
-	if (!p)
+	if (!q)
 		return;
 
 	ap_rputs("#EXTM3U\n", r);
 
-	for (q = p; q; q = q->next) {
+	for (; q; q = q->next) {
+		/* do not include file in playlist if streaming is not allowed */
+		if (!(q->flags & EF_ALLOWSTREAM))
+			continue;
+
 		ap_rprintf(r, "#EXTINF:%i,", q->length);
 		if (q->artist)
 			ap_rvputs(r, q->artist, " - ", NULL);
@@ -913,17 +878,17 @@ void send_playlist(request_rec *r, const mu_ent *const p, const mu_config *const
  * and http://feedvalidator.org/docs/rss2.html
  *
  * @param r Apache request_rec struct to handle connection details.
- * @param p A list of music entries.
+ * @param pack A pack of music entries.
  * @param conf MusicIndex configuration paramaters struct.
  */
-void send_rss(request_rec *r, const mu_ent *const p, const mu_config *const conf)
+void send_rss(request_rec *r, const mu_pack *const pack, const mu_config *const conf)
 {
-	const mu_ent *q;
+	const mu_ent *q = pack->head;
 	char date_buf[32];
 	struct tm time_buf;
 	unsigned short limit = conf->rss_items;
 
-	if (!p)
+	if (!q)
 		return;
 
 	ap_rputs("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
@@ -937,8 +902,8 @@ void send_rss(request_rec *r, const mu_ent *const p, const mu_config *const conf
 		"  <title>RSS Feed for ", ap_escape_html(r->pool, r->uri), "</title>\n"
 		"  <link>", NULL);
 	send_url(r, r->uri, NULL, conf, 1);
-	ap_rputs("</link>\n", r);
-	ap_rprintf(r, "  <description>%d most recent songs from %s</description>\n", conf->rss_items, ap_escape_html(r->pool, r->uri));
+
+	ap_rprintf(r, "</link>\n  <description>%d most recent songs from %s</description>\n", conf->rss_items, ap_escape_html(r->pool, r->uri));
 
 	if ((conf->options & MI_PODCAST) != 0)
 		ap_rprintf(r, "  <itunes:summary>%d most recent songs from %s</itunes:summary>\n", conf->rss_items, ap_escape_html(r->pool, r->uri));
@@ -948,7 +913,7 @@ void send_rss(request_rec *r, const mu_ent *const p, const mu_config *const conf
 		"  <ttl>60</ttl>\n",
 		r);
 
-	for (q = p; limit && q; q = q->next) {
+	for (; limit && q; q = q->next) {
 		if (q->filetype < 0)	/* XXX try to get rid of this test */
 			continue;
 
@@ -979,37 +944,29 @@ void send_rss(request_rec *r, const mu_ent *const p, const mu_config *const conf
 							q->length / 60, q->length % 60);
 			}
 		}
-		ap_rputs("   <description>\n", r);
-		ap_rputs("    Artist | Album | Track | Disc | Length | Genre | Bitrate | Freq | Type | Size\n", r);
-		ap_rprintf(r, "    %s | ", q->artist ? ap_escape_html(r->pool, q->artist) : "");
-		ap_rprintf(r, "%s | ", q->album ? ap_escape_html(r->pool, q->album) : "");
+
+		ap_rputs("   <description>\n    Artist | Album | Track | Disc | Length | Genre | Bitrate | Freq | Type | Size\n", r);
+		ap_rprintf(r, "    %s | %s | ", q->artist ? ap_escape_html(r->pool, q->artist) : "",
+				q->album ? ap_escape_html(r->pool, q->album) : "");
 		if (q->track)
-			ap_rprintf(r, "%u", q->track);
-		ap_rputs(" | ", r);
+			ap_rprintf(r, "%u | ", q->track);
 		if (q->posn)
-			ap_rprintf(r, "%u", q->posn);
-		ap_rputs(" | ", r);
+			ap_rprintf(r, "%u | ", q->posn);
 		if (q->length)
-			ap_rprintf(r, "%u:%.2u", q->length / 60, q->length % 60);
-		ap_rputs(" | ", r);
+			ap_rprintf(r, "%u:%.2u | ", q->length / 60, q->length % 60);
 		ap_rprintf(r, "%s | ", q->genre ? ap_escape_html(r->pool, q->genre) : "");
 		if (q->bitrate)
-			ap_rprintf(r, "%lu %s", q->bitrate >> 10, (q->flags & EF_VBR) ? "VBR" : "");
-		ap_rputs(" | ", r);
+			ap_rprintf(r, "%lu %s | ", q->bitrate >> 10, (q->flags & EF_VBR) ? "VBR" : "");
 		if (q->freq)
-			ap_rprintf(r, "%u", q->freq);
-		ap_rputs(" | ", r);
-		ap_rprintf(r, "%s", q->filetype ? filetype[q->filetype % FT_MAX].nametype : "");
-		ap_rputs(" | ", r);
-		ap_rprintf(r, "%lu\n", q->size);
+			ap_rprintf(r, "%u | ", q->freq);
+		ap_rprintf(r, "%s | %lu\n", q->filetype ? filetype[q->filetype % FT_MAX].nametype : "", q->size);
 		ap_rputs("   </description>\n"
 			"  </item>\n", r);
-		
+
 		limit--;
 	}
 	ap_rputs(" </channel>\n"
 		"</rss>\n", r);
-
 }
 
 /**
@@ -1053,7 +1010,7 @@ void send_randomdir(request_rec *r, const mu_config *const conf)
 		while ((dirent = readdir(dir)))
 			if ((dirent->d_name[0] != '.') && (access(r->filename, R_OK|X_OK) == 0))
 				nb++;
-		
+
 		/* once it's done, go back to the begining of the dir stream */
 		rewinddir(dir);
 		
@@ -1065,14 +1022,14 @@ void send_randomdir(request_rec *r, const mu_config *const conf)
 		while((i <= test) && (dirent = readdir(dir)))
 			if ((dirent->d_name[0] != '.') && (access(r->filename, R_OK|X_OK) == 0))
 				i++;
-			
+
 		/* construct a new filename with the current dirent */
 		nextfile = ap_pstrcat(r->pool, filename, dirent->d_name, NULL);
 		nexturi = ap_pstrcat(r->pool, uri, dirent->d_name, NULL);
 
 		/* done with the dir listing, close it */
 		closedir(dir);
-		
+
 		/* if the new filename is a directory, start the loop again, else, break */
 		if (!ap_is_directory(r->pool, nextfile)) {
 			/* XXX this is probably not good either */
@@ -1099,30 +1056,52 @@ void send_randomdir(request_rec *r, const mu_config *const conf)
  */
 void send_foot(request_rec *r, const mu_config *const conf)
 {
+	char *icn = NULL;
+	request_rec *subreq = ap_sub_req_lookup_uri(conf->directory, r, NULL);
+
 	ap_rputs("<!-- begin footer -->\n"
 		"<!-- mod_musicindex v." MUSIC_VERSION_STRING " -->\n"
-		"<!-- Authors: " MUSIC_AUTHORS_STRING " -->\n", r);
-	ap_rputs("<div id=\"footer\">\n"
+		"<!-- Authors: " MUSIC_AUTHORS_STRING " -->\n"
+		"<div id=\"footer\">\n"
 		" <div id=\"valid\">\n", r);
-	/* Should be possible to use on-site validation icons */
-	ap_rputs("  <a href=\"http://validator.w3.org/check?uri=referer\">\n"
-		"   <img src=\"http://www.w3.org/Icons/valid-xhtml11\"\n"
-        	"    alt=\"Valid XHTML 1.1!\" height=\"31\" width=\"88\" />\n"
-		"  </a>\n", r);
-	ap_rputs("  <a href=\"http://jigsaw.w3.org/css-validator/check/referer\">\n"
-		"   <img src=\"http://jigsaw.w3.org/css-validator/images/vcss\"\n"
-		"    alt=\"Valid CSS!\" height=\"31\" width=\"88\" />\n"
-		"  </a>\n", r);
-	ap_rputs("  <a href=\"http://rss.scripting.com/\">\n"
-		"   <img src=\"http://cyber.law.harvard.edu/blogs/gems/tech/loveRss.gif\"\n"
-		"    width=\"65\" height=\"59\" alt=\"Valid RSS feed.\" />\n"
-		"  </a>\n", r);
+	/* if we have local icons, use them */
+	/* valid xhtml 1.1 */
+	if (subreq && !access(ap_pstrcat(r->pool, subreq->filename, "/valid-xhtml11", NULL), R_OK)) {
+		icn = ap_pstrcat(r->pool, conf->directory, "/valid-xhtml11", NULL);
+		ap_rvputs(r, "  <img src=\"", icn, "\" alt=\"Valid XHTML 1.1!\" />\n", NULL);
+	}
+	else
+		ap_rputs("  <a href=\"http://validator.w3.org/check?uri=referer\">\n"
+			 "   <img src=\"http://www.w3.org/Icons/valid-xhtml11\"\n"
+			 "    alt=\"Valid XHTML 1.1!\" height=\"31\" width=\"88\" />\n"
+			 "  </a>\n", r);
+	/* valid CSS */
+	if (subreq && !access(ap_pstrcat(r->pool, subreq->filename, "/vcss", NULL), R_OK)) {
+		icn = ap_pstrcat(r->pool, conf->directory, "/vcss", NULL);
+		ap_rvputs(r, "  <img src=\"", icn, "\" alt=\"Valid CSS!\" />\n", NULL);
+	}
+	else
+		ap_rputs("  <a href=\"http://jigsaw.w3.org/css-validator/check/referer\">\n"
+			 "   <img src=\"http://jigsaw.w3.org/css-validator/images/vcss\"\n"
+			 "    alt=\"Valid CSS!\" height=\"31\" width=\"88\" />\n"
+			 "  </a>\n", r);
+	/* Valid RSS */
+	if (subreq && !access(ap_pstrcat(r->pool, subreq->filename, "/loveRss.gif", NULL), R_OK)) {
+		icn = ap_pstrcat(r->pool, conf->directory, "/loveRss.gif", NULL);
+		ap_rvputs(r, "  <img src=\"", icn, "\" alt=\"Valid RSS!\" />\n", NULL);
+	}
+	else
+		ap_rputs("  <a href=\"http://rss.scripting.com/\">\n"
+			 "   <img src=\"http://cyber.law.harvard.edu/blogs/gems/tech/loveRss.gif\"\n"
+			 "    width=\"65\" height=\"59\" alt=\"Valid RSS feed.\" />\n"
+			 "  </a>\n", r);
+	/* Signature */
 	ap_rputs(" </div>\n"
 		" <div id=\"name\">"
 		"<a href=\"http://freshmeat.net/projects/musicindex/\">"
 		"MusicIndex v." MUSIC_VERSION_STRING "</a></div>\n"
-		"</div>\n", r);
-	ap_rputs("<!-- end footer -->\n\n"
+		"</div>\n"
+		"<!-- end footer -->\n\n"
 		"</body>\n"
 		"</html>", r);
 }
