@@ -2,11 +2,11 @@
  *  mod_musicindex.c
  *  mod_musicindex
  *
- *  $Id: mod_musicindex.c 662 2006-07-09 21:05:07Z varenet $
+ *  $Id: mod_musicindex.c 756 2007-05-01 19:27:56Z varenet $
  *
  *  Created by Regis BOUDIN on Sat Dec 28 2002.
  *  Copyright (c) 2002-2006 Regis BOUDIN
- *  Copyright (c) 2002-2005 Thibaut VARENE
+ *  Copyright (c) 2002-2005,2007 Thibaut VARENE
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License version 2.1,
@@ -18,10 +18,10 @@
 /**
  * @mainpage
  * @section about About
- *	mod_musicindex is an Apache module aimed at being a C implementation
+ *	mod_musicindex is an Apache module initially aimed at being a C implementation
  *	of the Perl module <a href="http://search.cpan.org/dist/Apache-MP3/">
  *	Apache::MP3</a>. It allows nice displaying of directories containing
- *	MP3, Ogg Vorbis, FLAC or MP4 files, including sorting them on various
+ *	MP3, Ogg Vorbis, FLAC or MP4/AAC files, including sorting them on various
  *	fields, streaming/downloading them, constructing playlist, and searching.
  *
  * @section features Features
@@ -66,7 +66,7 @@
  *	Developers should also take special care to typing. Strict typing is the
  *	first protection against coding mistakes that might lead to dreadful
  *	crashes. The same applies to function modifier (in particular the
- *	<code>static</code> modifier, that should be used whenever appropriate.
+ *	@c static modifier, that should be used whenever appropriate.
  *
  *	Finally, we try to put Apache's specific arguments to functions at the
  *	beginning of their param list.
@@ -93,9 +93,9 @@
  *	- As a rule of thumb, always try to declare all function variables at the
  *	beginning of the function, and avoid as much as possible to declare
  *	subvariables in blocks.
- *	- In any case, <b>never</b> declare new variables <b>after</b> code in a block.
+ *	- In any case, @b never declare new variables @b after code in a block.
  *	- Finally, special care should be bore to storage space: always declare
- *	as <i>little</i> as possible: we want the module to consume as few
+ *	as @e little as possible: we want the module to consume as few
  *	memory as possible. That also applies to functions return types.
  */
  
@@ -106,23 +106,22 @@
  * This file is the core of the module. It contains Apache's mandatory stuff.
  *
  * @author Regis Boudin
- * @version $Revision: 662 $
+ * @version $Revision: 756 $
  * @date 2002-2005
+ * @date 2007
  *
  * http://httpd.apache.org/docs-2.0/developer/API.html
  * http://httpd.apache.org/dev/apidoc/
+ * http://www.apachetutor.org/dev/pools
  *
  * @warning Use it at your own risks!
+ * @warning sub requests are @b EXTREMELY cpu expensive
  *
  * @todo Complete code documentation.
  * @todo (low pri whishlist) On the fly tag rewriting or metadata support.
- * @todo Considering the previous item, we might also need a handler for
- * streaming, otherwise we won't be able to restrain file downloading
- * (Apache::MP3 doesn't do anything better there fwiw).
  * @todo prepare the possibility to generate ices/mpd playlists.
- * @todo i18n (in progress).
  * @todo enforce strict type policy.
- * @todo get rid of str*cmp() whenever possible.
+ * @todo get rid of str*cmp() whenever possible and eventually implement "likely()".
  * @todo review the randomdir implementation.
  *
  * @todo redesign cache integration to speed it up:
@@ -139,8 +138,7 @@
  *
  * @todo Audit the code: in many places we are using a lot of arguments passed
  * 	to functions that could easily be gotten otherwise by dereferencing a 
- * 	few pointers. Try to figure out what's the most efficient (probably
- * 	passing a lot of args), balancing with storage space usage.
+ * 	few pointers. Passing too many args increases stack usage.
  *
  * @todo Decide how long we want to support Apache 1.3. APR provides lots of
  *	useful features, and the API compatibility between 1.3 and 2.2 is not
@@ -159,6 +157,9 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#ifdef HAVE_SETLOCALE
+#include <locale.h>
+#endif
 
 #define MUSIC_HEADER_STRING "mod_musicindex/" MUSIC_VERSION_STRING
 
@@ -167,8 +168,12 @@
 static const char *handlers[] = {
 	"audio/mpeg",
 	"application/ogg",
-	"audio/flac",
+	"audio/ogg",
 	"audio/x-ogg", /* At some point, we should be able to remove this one */
+	"audio/flac",
+	"audio/x-flac",
+	"application/x-flac",
+	"audio/mp4",
 	NULL
 };
 #else
@@ -176,14 +181,22 @@ static const char *handlers[] = {
 #endif
 
 #ifdef ENABLE_OUTPUT_ARCHIVE
-extern void send_tarball(request_rec *r, const mu_ent *const p, const mu_config *const conf);
+extern void send_tarball(request_rec *, const mu_pack *const, const mu_config *const);
 #endif
 
 static int handle_musicindex(request_rec *r)
 {
 	/* Get the module configuration */
 	mu_config *const conf = (mu_config *)ap_get_module_config(r->per_dir_config, &musicindex_module);
-	mu_ent *main_list = NULL, *custom_list = NULL;
+	mu_pack master_pack = {
+		.head = NULL,
+		.dirnb = 0,
+		.filenb = 0,
+	},	custom_pack = {
+		.head = NULL,
+		.dirnb = 0,
+		.filenb = 0,
+	};
 
 	/* if the the module is not active, or if request is not a GET or a POST,
 	   let the dir to another module */
@@ -255,17 +268,17 @@ static int handle_musicindex(request_rec *r)
 		((conf->options & MI_DWNLDRQ) == MI_DWNLDLST) ||
 		((conf->options & MI_DWNLDRQ) == MI_COOKIEDWNLD)) {
 		/* We build a custom playlist, don't bother with current directory */
-		main_list = build_custom_list(r, conf);
+		build_custom_list(r, &master_pack, conf);
 	}
 	else {
 		/* Build current directory content (recursive if necessary) */
-		main_list = make_music_entry(r, r->pool, NULL, conf, NULL, MI_RECURSIVE);
-		main_list = quicksort(main_list, NULL, conf);
+		make_music_entry(r, r->pool, &master_pack, conf, NULL, MI_RECURSIVE);
+		listsort(&master_pack, conf->order);
 	}
 
 	/* Build custom link list, from conf->custom_list string */
 	if (((conf->options & MI_STREAM) == 0) && (conf->custom_list != NULL))
-		custom_list = build_custom_list(r, conf);
+		build_custom_list(r, &custom_pack, conf);
 
 	/* Depending on the request type, set the HTTP headers. By default we
 	   return a web page. */
@@ -293,26 +306,26 @@ static int handle_musicindex(request_rec *r)
 	ap_send_http_header(r);
 
 	if (r->header_only)
-		return 0;
+		return OK;
 
 	if (conf->options & MI_STREAM)
-		send_playlist(r, main_list, conf);
+		send_playlist(r, &master_pack, conf);
 	else if (conf->options & MI_RSS)
-		send_rss(r, main_list, conf);
+		send_rss(r, &master_pack, conf);
 #ifdef ENABLE_OUTPUT_ARCHIVE
 	else if (conf->options & MI_TARBALL)
-		send_tarball(r, main_list, conf);
+		send_tarball(r, &master_pack, conf);
 #endif
 	else {
 		send_head(r, conf);
 		if (conf->search == NULL)
-			send_directories(r, main_list, conf);
-		send_tracks(r, main_list, conf);
-		send_customlist(r, custom_list, conf);
+			send_directories(r, &master_pack, conf);
+		send_tracks(r, &master_pack, conf);
+		send_customlist(r, &custom_pack, conf);
 		send_foot(r, conf);
 	}
 
-	return 0;
+	return OK;
 }
 
 /**
@@ -336,10 +349,7 @@ static int handle_musicfile(request_rec *r)
 		return DECLINED;
 
 #ifdef BUILD_FOR_APACHE2
-	for (i=0; handlers[i] && strcmp(r->handler, handlers[i]); i++) {
-		if (!strcmp(r->handler, handlers[i]))
-			break;
-	}
+	for (i=0; handlers[i] && strcmp(r->handler, handlers[i]); i++);
 
 	if (handlers[i] == NULL)
 		return DECLINED;
@@ -351,7 +361,7 @@ static int handle_musicfile(request_rec *r)
 		- Stream is allowed and no icecast server is set
 	   Otherwise, we consider it is a forbidden request. Keep in mind that
 	   DECLINED just means the module won't handle the request, whereas
-	   HTTP_FORBIDDEN means the request is not allowed, not matter what. */
+	   HTTP_FORBIDDEN means the request is not allowed, no matter what. */
 	/* XXX We should handle fds by ourselves, and set the mime type according
 	   to our standards, maybe? */
 	if (r->args == NULL) {
@@ -370,7 +380,11 @@ static int handle_musicfile(request_rec *r)
 
 
 	if ((conf->options & MI_ALLOWSTREAM) && (strcmp(r->args, "stream") == 0)) {
-		mu_ent	*head = NULL;
+		mu_pack master_pack = {
+			.head = NULL,
+			.dirnb = 0,
+			.filenb = 0,
+		};
 
 		ap_set_content_type(r, "audio/x-mpegurl");
 		ap_table_setn(r->headers_out, "Content-Disposition",
@@ -379,12 +393,12 @@ static int handle_musicfile(request_rec *r)
 		ap_send_http_header(r);
 
 		if (r->header_only)
-			return 0;
+			return OK;
 
-		head = make_music_entry(r, r->pool, head, conf, NULL, 0);
-		send_playlist(r, head, conf);
+		make_music_entry(r, r->pool, &master_pack, conf, NULL, 0);
+		send_playlist(r, &master_pack, conf);
 
-		return 0;
+		return OK;
 	}
 
 	return HTTP_FORBIDDEN;
@@ -401,8 +415,11 @@ static int handle_musicfile(request_rec *r)
 static void musicindex_init(server_rec *s, apr_pool_t *p)
 {
 #ifdef HAVE_GETTEXT
+#ifdef HAVE_SETLOCALE
 	setlocale(LC_ALL, "");
+#endif
 	textdomain("mod_musicindex");
+	bind_textdomain_codeset("mod_musicindex","UTF-8");	/* forcons la raie */
 #endif
 
 	ap_add_version_component(MUSIC_HEADER_STRING);
@@ -412,9 +429,12 @@ static void musicindex_init(server_rec *s, apr_pool_t *p)
 static const handler_rec musicindex_handlers[] = {
 	{DIR_MAGIC_TYPE, handle_musicindex},
 	{"audio/mpeg", handle_musicfile},
+	{"audio/ogg", handle_musicfile},
 	{"audio/x-ogg", handle_musicfile}, /* Backward compatibility */
 	{"application/ogg", handle_musicfile},
 	{"audio/flac", handle_musicfile},
+	{"audio/x-flac", handle_musicfile},
+	{"application/x-flac", handle_musicfile},
 	{"audio/mp4", handle_musicfile},	/* XXX cette struct sert a quoi exactement? MP4 marchait meme sans ca */
 	{NULL}
 };
@@ -455,8 +475,11 @@ module MODULE_VAR_EXPORT musicindex_module = {
 static int musicindex_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
 {
 #ifdef HAVE_GETTEXT
+#ifdef HAVE_SETLOCALE
 	setlocale(LC_ALL, "");
+#endif
 	textdomain("mod_musicindex");
+	bind_textdomain_codeset("mod_musicindex","UTF-8");	/* forcons la raie */
 #endif
 
 	ap_add_version_component(p, MUSIC_HEADER_STRING);
