@@ -2,11 +2,11 @@
  *  mod_musicindex.h
  *  mod_musicindex
  *
- *  $Id: mod_musicindex.h 678 2006-10-14 19:49:51Z varenet $
+ *  $Id: mod_musicindex.h 759 2007-05-05 17:18:09Z varenet $
  *
  *  Created by Thibaut VARENE on Thu Mar 20 2003.
  *  Copyright (c) 2003-2006 Regis BOUDIN
- *  Copyright (c) 2003-2005 Thibaut VARENE
+ *  Copyright (c) 2003-2007 Thibaut VARENE
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License version 2.1,
@@ -47,13 +47,11 @@
 #define MUSIC_AUTHORS_STRING	"R. Boudin & T. Varene"	/**< Authors */
 
 /*
- * Compile time configuration option
+ * Compile time configuration options
  */
 #undef NO_TITLE_STRIP	/* Disable pretty printing for untagged files */
 #undef NO_BR_NOMINAL	/* Disable display of nominal bitrate when available */
-#undef SHOW_THUMBNAILS	/* Enable custom folder icon in folder listings */
-#define NEW_HEADER	/**< Enable new header display */
-
+#define PRETTY_FOLDERS 1	/**< Define to 0 to disable per subdirectory icons */
 
 /* these are not defined on BSD */
 #ifndef FALSE
@@ -64,30 +62,35 @@
 /* a very common macro */
 #define ARRAY_SIZE(x) (sizeof(x) / sizeof(x[0]))
 
-/*
- * Various constants used almost everywhere
+/**
+ * Enumerates sort options in a more elegant way than using dozens of defines.
+ * This enum should be kept in sync with the sort_order_functions[] array in sort.c,
+ * otherwise it's all about games.
  */
-#define SB_ALBUM	'A'		/**< Sort by album */
-#define SB_ARTIST	'B'		/**< Sort by artist */
-#define SB_DATE		'D'		/**< Sort by date */
-#define SB_GENRE	'G'		/**< Sort by genre */
-#define SB_FREQ		'E'		/**< Sort by frequency (samplerate) */
-#define SB_FILETYPE	'F'		/**< Sort by filetype */
-#define SB_LENGTH	'L'		/**< Sort by length */
-#define SB_FILENAME	'M'		/**< Sort by filename */
-#define SB_TRACK	'N'		/**< Sort by track */
-#define SB_POSN		'P'		/**< Sort by Part Of Set nb */
-#define SB_MTIME	'Q'		/**< Sort by modification time */
-#define SB_BITRATE	'R'		/**< Sort by bitrate */
-#define SB_SIZE		'S'		/**< Sort by filesize */
-#define SB_TITLE	'T'		/**< Sort by title */
-#define SB_URI		'U'		/**< Sort by URI */
-#define SB_DIR		'V'		/**< Sort by directory */
-#define SB_DEFAULT	'U'		/**< Default sort order (used in search results) */
-#define SB_RANDOM	'Z'		/**< Sort randomly */
+enum {
+	SB_NULL,		/**< Avoid 0 value which has all sorts of nasty side effects */
+	SB_ALBUM,		/**< Sort by album */
+	SB_POSN,		/**< Sort by Part Of Set nb */
+	SB_TRACK,		/**< Sort by track */
+	SB_ARTIST,		/**< Sort by artist */
+	SB_TITLE,		/**< Sort by title */
+	SB_LENGTH,		/**< Sort by length */
+	SB_BITRATE,		/**< Sort by bitrate */
+	SB_FREQ,		/**< Sort by frequency (samplerate) */
+	SB_FILETYPE,		/**< Sort by filetype */
+	SB_FILENAME,		/**< Sort by filename */
+	SB_URI,			/**< Sort by URI */
+	SB_GENRE,		/**< Sort by genre */
+	SB_DATE,		/**< Sort by date */
+	SB_SIZE,		/**< Sort by filesize */
+	SB_MTIME,		/**< Sort by modification time */
+	SB_RANDOM,		/**< Sort randomly */
+	SB_DIR,			/**< Sort by directory */
+	SB_MAX,			/**< Number of elements in this enum */
+};
 
-#define ARG_NUMBER	18		/**< Number of args to sort lists */
 
+#define SB_DEFAULT	SB_URI	/**< Default sort order (used in search results) */
 
 /*
  * Flags used in mu_config struct, fields "options" and "options_not".
@@ -214,9 +217,16 @@
  #define ap_psprintf apr_psprintf
  #define ap_pstrndup apr_pstrndup
  #define ap_base64decode_len apr_base64_decode_len
- #define ap_http_method ap_http_scheme
  #define ap_base64decode apr_base64_decode
 
+ #if defined(AP20)
+  /* nada */
+ #elif defined(AP22)
+  #define ap_http_method ap_http_scheme
+ #else
+  #error "unknown apache API"
+ #endif
+
  #include <apr_strings.h>
  #include <apr_base64.h>
  #include <ap_compat.h>
@@ -255,29 +265,61 @@ typedef struct mu_ent {
 	unsigned long bitrate;	/**< bitrate in bits per second */
 	unsigned long size;	/**< size in bytes */
 	unsigned long mtime;	/**< last modification time */
-	const char *file;	/**< file path */
-	const char *filename;	/**< file path */
+	const char *file;	/**< file uri. Relative uri unless MI_CUSTOM. */
+	const char *filename;	/**< filename (as in uri). file absolute *path* if MI_TARBALL */
 	const char *uri;	/**< file URI (unique) */
 	const char *album;	/**< album (if any) */
 	const char *artist;	/**< artist (if any) */
 	const char *title;	/**< title (as shown on the webpages) */
 	const char *genre;	/**< genre (type of music - if any) */
-	struct mu_ent *next;
+	const struct mu_ent *next;
 } mu_ent;
 
+typedef struct {
+	const mu_ent *head;
+	unsigned long filenb;
+	unsigned short dirnb;
+} mu_pack;
+
 typedef struct mu_config mu_config;
 
+/**
+ * Cache backend structure, defining a standard set of operations that all
+ * cache backends must implement.
+ */
 typedef struct {
-	void* (*opendir)(request_rec *r, mu_ent **entries, const mu_config *const conf, const mu_ent_names * const names, unsigned long soptions);
+	/** This function is the entry point into the cache backend system.
+	    It can be used in many weird ways, but is mostly expected to either
+	    return NULL if the directory backend can't provide a full directory
+	    listing, and otherwise populate the entries with whatever is in the
+	    cache. If return is !NULL, (cache_backend::readdir()) will be
+	    subsequently called on the provided return pointer.
+	    See musicindex_opendir() for details. */
+	void* (*opendir)(request_rec *r, mu_pack *pack, const mu_config *const conf,
+			const mu_ent_names * const names, unsigned long soptions);
+
+	/** If the above call provided a dir handler, this function will be used.
+	    Its primary intent is to provide a way for the cache backend to let
+	    the program know that some files couldn't be processed at cache level,
+	    and that they need to be parsed by the standard playlist system.
+	    Each consecutive call to this function should return the name of a
+	    file in the current directory to be processed, terminating with NULL,
+	    a la readdir. The only nitpick here is that files listed here will
+	    still hit make_cache_entry(), that is the following cache_backend::make_entry() */
 	const char* (*readdir)(void *dir, const mu_config *const conf);
+
+	/** At the end of directory processing, this function is called, still
+	    mimicing opendir/readdir/closedir. */
 	void (*closedir)(void *dir, const mu_config *const conf);
-	mu_ent* (*make_entry)(request_rec *r, apr_pool_t *pool, FILE *const in, mu_ent *const head, const mu_config *const conf, mu_ent_names *const names);
+	
+	/** This call reads cached metadata for a given file */
+	mu_ent* (*make_entry)(request_rec *r, apr_pool_t *pool, FILE *const in,
+		const mu_config *const conf, mu_ent_names *const names);
+
+	/** This call tells the cache backend to write metadata for a given file. */
 	void (*write)(request_rec *r, mu_ent *entry, const mu_config *const conf, const mu_ent_names * const names);
 } cache_backend;
 
-/** sort function pointer type (see inf.c) */
-typedef short (*inf_ptr)(const mu_ent *const, const mu_ent *const);
-
 /** Configuration handling structure. We use it to share configuration
  * throughout the whole module */
 struct mu_config {
@@ -287,9 +329,8 @@ struct mu_config {
 	unsigned short options;		/**< diverse options */
 	unsigned short options_not;	/**< used to disable options */
 
-	char order[ARG_NUMBER];		/**< Sort order */
-	char fields[ARG_NUMBER];	/**< Fields to display */
-	inf_ptr order_functions[ARG_NUMBER];	/**< Functions array for sort */
+	unsigned char order[SB_MAX+1];	/**< Sort order. */
+	unsigned char fields[SB_MAX+1];	/**< Fields to display */
 
 	const char *title;		/**< title of the page */
 
