@@ -2,11 +2,11 @@
  *  playlist.c
  *  mod_musicindex
  *
- *  $Id: playlist.c 670 2006-07-31 21:34:34Z varenet $
+ *  $Id: playlist.c 754 2007-04-29 00:57:53Z varenet $
  *
  *  Created by Thibaut VARENE on Thu Mar 20 2003.
  *  Copyright (c) 2003-2004 Regis BOUDIN
- *  Copyright (c) 2003-2006 Thibaut VARENE
+ *  Copyright (c) 2003-2007 Thibaut VARENE
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License version 2.1,
@@ -20,8 +20,8 @@
  *
  * @author Regis Boudin
  * @author Thibaut Varene
- * @version $Revision: 670 $
- * @date 2003-2006
+ * @version $Revision: 754 $
+ * @date 2003-2007
  *
  * This is where the magic takes place. Here are the functions in charge of
  * analyzing the content of the folders requested by the client, and of
@@ -72,20 +72,19 @@ const struct ftype const filetype[] = {
 /**
  * Fallback function if the file could not be recognised.
  *
+ * @param r Apache request_rec struct
  * @param pool Pool
- * @param head Head
  * @param in file to parse (closed on normal exit)
  * @param conf MusicIndex configuration paramaters struct
  * @param names Names
- * @param r Apache request_rec struct
  *
  * @return head, whatever happens.
  */
-static mu_ent *make_no_entry(request_rec *r, apr_pool_t *pool, FILE *const in,
-	mu_ent *const head, const mu_config *const conf, mu_ent_names *const names)
+static inline mu_ent *make_no_entry(request_rec *r, apr_pool_t *pool, FILE *const in,
+	const mu_config *const conf, mu_ent_names *const names)
 {
 	fclose(in);
-	return head;
+	return NULL;
 }
 
 /**
@@ -94,61 +93,62 @@ static mu_ent *make_no_entry(request_rec *r, apr_pool_t *pool, FILE *const in,
  * @param r Apache request_rec struct
  * @param pool Apache memory pool
  * @param in file to parse (closed on normal exit)
- * @param head head list entry
  * @param conf MusicIndex configuration parameters
  * @param names Names struct associated with the current in file
  *
  * @return the cache data on success, head otherwise
  */
 static mu_ent *make_cache_entry(request_rec *r, apr_pool_t *pool, FILE *const in,
-	mu_ent *const head, const mu_config *const conf, mu_ent_names *const names)
+	const mu_config *const conf, mu_ent_names *const names)
 {
-	if (conf->cache != NULL) {
-		if (conf->cache->make_entry != NULL) {
-			mu_ent *result = conf->cache->make_entry(r, pool, in, head, conf, names);
-			if (result != NULL) {
-				result->flags |= EF_INCACHE;
-				return result;
-			}
+	if (conf->cache && conf->cache->make_entry) {
+		mu_ent *result = conf->cache->make_entry(r, pool, in, conf, names);
+		if (result) {
+			result->flags |= EF_INCACHE;
+			return result;
 		}
 	}
 	
-	return head;
+	return NULL;
 }
 
 /**
  * Function pointers array.
  *
  * Of course, as we defined function pointers, we will use them in a nice
- * array. They will be called in a sequential manner.<br>
- * make_cache_entry() <b>MUST</b> be called first.<br>
- * <i>Note</i> : a make_dir_entry() function to handle directories has
+ * array. They will be called in a sequential manner. @n
+ * make_cache_entry() @b MUST be called first. @n
+ * @e Note : a make_dir_entry() function to handle directories has
  * already been tried, unfortunately, it implies some bad side effects.
  * Plus, it probably does not make things any quicker and increases the
  * number of function calls. (that's a personnal NB to the developpers,
- * to avoid repeating twice the same havoc as time goes ;o)
+ * to avoid repeating twice the same havoc as time goes by ;o)
+ *
+ * Functions are ordered in what we believe is a top-down order of most frequent
+ * filetypes.
  */
-static make_entry_ptr make_entry[] = {
+static const make_entry_ptr const make_entry[] = {
 	make_cache_entry,
-#ifdef	ENABLE_CODEC_VORBIS
-	make_ogg_entry,
-#endif
 #ifdef	ENABLE_CODEC_MP3
 	make_mp3_entry,
 #endif
-#ifdef	ENABLE_CODEC_FLAC
-	make_flac_entry,
+#ifdef	ENABLE_CODEC_VORBIS
+	make_ogg_entry,
 #endif
 #ifdef	ENABLE_CODEC_MP4
 	make_mp4_entry,
+#endif
+#ifdef	ENABLE_CODEC_FLAC
+	make_flac_entry,
 #endif
 	make_no_entry,	/* MUST be last before NULL */
 	NULL
 };
 
+/** Wrapper struct for the cache subsystem */
 typedef struct mu_dir {
 	void *dir;
-	unsigned char handled;
+	unsigned char handled;	/**< dir is handled by the cache */
 } mu_dir;
 
 /**
@@ -161,30 +161,25 @@ typedef struct mu_dir {
  * with. These files names will be retrieved with musicindex_readdir().
  * We mimic the opendir()/readdir()/closedir() scheme.
  *
- * These functions stay here as static, so th compiler will happily inline them
- * for us.
- *
  * @param r the Apache request
  * @param dir Abstract pointer to our custom directory handler
- * @param entries list pointer (double indirection)
+ * @param pack pack pointer
  * @param conf Configuration of the current request
  * @param names structure containing the name of the directory
  * @param soptions the current extended options
  *
  * @return handler for the directory.
  */
-static void 
-musicindex_opendir(request_rec *r, mu_dir *dir, mu_ent **entries, const mu_config *const conf,
+static inline void 
+musicindex_opendir(request_rec *r, mu_dir *dir, mu_pack *const pack, const mu_config *const conf,
 			const mu_ent_names * const names, unsigned long soptions)
 {
 	/* It's all we need to initialise. Don't add a call to memset */
 	dir->dir = NULL;
 
 	/* A cache is configured, try to read directly data from it */
-	if (conf->cache != NULL) {
-		if (conf->cache->opendir != NULL)
-			dir->dir = conf->cache->opendir(r, entries, conf, names, soptions);
-	}
+	if (conf->cache && conf->cache->opendir)
+		dir->dir = conf->cache->opendir(r, pack, conf, names, soptions);
 	
 	/* Cache data could not be used, fall back to the old opendir way */
 	if (dir->dir == NULL) {
@@ -205,24 +200,23 @@ musicindex_opendir(request_rec *r, mu_dir *dir, mu_ent **entries, const mu_confi
  * @return pointer to the string of the filename, or NULL if there is nothing
  *         left to open.
  */
-static const char *musicindex_readdir(mu_dir *dir, const mu_config *const conf)
+static inline const char *musicindex_readdir(mu_dir *dir, const mu_config *const conf)
 {
 	if (dir->handled == 0 ) {
 		struct dirent *dstruct = NULL;
 	
 		dstruct = readdir((DIR *)(dir->dir));
 
-		if ( NULL == dstruct)
+		if (NULL == dstruct)
 			return NULL;
 		else
 			return dstruct->d_name;
 
 	}
 	else {
-		if (conf->cache != NULL) {
-			if (conf->cache->readdir != NULL)
-				return conf->cache->readdir(dir->dir, conf);
-		}
+		/* not all cache subsys can perform whole directory read */
+		if (conf->cache && conf->cache->readdir)
+			return conf->cache->readdir(dir->dir, conf);
 	}
 	return NULL;
 }
@@ -235,7 +229,7 @@ static const char *musicindex_readdir(mu_dir *dir, const mu_config *const conf)
  *
  * @return nothing.
  */
-static void musicindex_closedir(mu_dir *dir, const mu_config *const conf)
+static inline void musicindex_closedir(mu_dir *dir, const mu_config *const conf)
 {
 	/* A cache is configured, try to read directly data from it */
 	if (dir->handled == 0)
@@ -249,39 +243,31 @@ static void musicindex_closedir(mu_dir *dir, const mu_config *const conf)
  *
  * @param r Apache current request
  * @param conf MusicIndex configuration paramaters struct
- * @param names Names
+ * @param local_options This function can be called in subdirectories for which
+ *	the configuration can differ from the current dir's one. This paramater
+ *	allows the caller to specify the options to check against.
  *
- * @return 0 if .
+ * @return TRUE if directory descend is allowed, FALSE otherwise.
  */
-static int go_through_directory(request_rec *r, const mu_config *const conf, mu_ent_names *names)
+static inline short
+go_through_directory(request_rec *r, const mu_config *const conf, const unsigned short local_options)
 {
-	request_rec *sub_req = NULL;
-	unsigned short local_options;
-
-	/* When going into subdir, get their local options first */
-	sub_req = ap_sub_req_lookup_uri(names->uri, r, NULL);
-	if (sub_req == NULL)
-		return -1;
-
-	local_options = ((mu_config*)ap_get_module_config(sub_req->per_dir_config, &musicindex_module))->options;
-	ap_destroy_sub_req(sub_req);
-
 	/* before dealing with a directory, sanity checks */
 	/* First, is the module enabled ? */
 	if (!(local_options & MI_ACTIVE))
-		return -1;
+		return FALSE;
 
 	/* playall... is stream allowed ? */
 	if (((conf->options & MI_STREAMALL) == MI_STREAMALL)
 		&& !(local_options & MI_ALLOWSTREAM))
-		return -1;
+		return FALSE;
 
 	/* Searching... Is searching allowed ? */
 	if ((conf->search) && !(local_options & MI_ALLOWSEARCH))
-		return -1;
+		return FALSE;
 
 	/* Everything is fine, go on */
-	return 0;
+	return TRUE;
 }
 
 /**
@@ -297,9 +283,12 @@ static int go_through_directory(request_rec *r, const mu_config *const conf, mu_
  * @bug I'm afraid if we're asked to recurse into too many subdirs, we might
  *	end up with too many open fdesc.
  *
+ * @note It'd be very nice to be able to make this function tail recursive,
+ * unfortunately it looks like it's going to be rather troublesome to do so.
+ *
  * @param pool Pool
  * @param r Apache request_rec struct to handle log writings (debugging)
- * @param head Head
+ * @param pack A pack of music entries to populate
  * @param conf MusicIndex configuration paramaters struct
  * @param names Names
  * @param soptions bitwise special options: these are made of the extended options flag set
@@ -307,10 +296,10 @@ static int go_through_directory(request_rec *r, const mu_config *const conf, mu_
  *
  * @return When possible, struct mu_ent correctly set up
  */
-mu_ent *make_music_entry(request_rec *r, apr_pool_t *pool,
-	mu_ent *const head, const mu_config *const conf, mu_ent_names *names, unsigned long soptions)
+void make_music_entry(request_rec *r, apr_pool_t *pool, mu_pack *const pack,
+	const mu_config *const conf, mu_ent_names *names, unsigned long soptions)
 {
-	mu_ent			*p = head;
+	mu_ent		*p = NULL;
 	register unsigned short	i;
 	char			*uri;
 
@@ -319,7 +308,7 @@ mu_ent *make_music_entry(request_rec *r, apr_pool_t *pool,
 		while parsed_uri.path is the path to the file on the webserver,
 		so no assertion can be made on their respective length */
 		if ((strlen(r->filename) >= MAX_STRING) || (strlen(r->parsed_uri.path) >= MAX_STRING))
-			return head;
+			return;
 		names = ap_palloc(pool, sizeof(mu_ent_names));
 		strcpy(names->filename, r->filename);
 		strcpy(names->uri, r->parsed_uri.path);
@@ -332,14 +321,14 @@ mu_ent *make_music_entry(request_rec *r, apr_pool_t *pool,
 		continue;
 
 	if (*(++uri) == '.')	/* we don't want "invisible" files to show */
-		return head;
+		return;
 
 	if (access(names->filename, R_OK) != 0)
-		return head;
+		return;
 
 	if ((ap_is_directory(pool, names->filename))) {
-
 		request_rec *subreq;
+		int local_options = 0;
 		char *fn = names->filename + strlen(names->filename) - 1;
 
 		if (*fn++ != '/')
@@ -351,40 +340,56 @@ mu_ent *make_music_entry(request_rec *r, apr_pool_t *pool,
 			*uri++ = '/';
 		*uri = '\0';
 
-		subreq = ap_sub_req_lookup_uri(names->uri, r, NULL);
+		if (PRETTY_FOLDERS) {
+			subreq = ap_sub_req_lookup_uri(names->uri, r, NULL);
 
-		soptions &= ~(MI_ALLOWFETCH|MI_ALLOWRSS);
-		
-		if (subreq != NULL) {
-			int local_options = ((mu_config *)ap_get_module_config(subreq->per_dir_config, &musicindex_module))->options;
-			int local_rss = ((mu_config *)ap_get_module_config(subreq->per_dir_config, &musicindex_module))->rss_items;
-			ap_destroy_sub_req(subreq);
-			
-			/* XXX pkoi pas soptions = local_options ? */
-			soptions |= (local_options & MI_ALLOWFETCH);
-			
-			if (local_rss > 0)
-				soptions |= MI_ALLOWRSS;
+			soptions &= ~(MI_ALLOWFETCH|MI_ALLOWRSS);
+
+			if (subreq != NULL) {
+				local_options = ((mu_config *)ap_get_module_config(subreq->per_dir_config, &musicindex_module))->options;
+				int local_rss = ((mu_config *)ap_get_module_config(subreq->per_dir_config, &musicindex_module))->rss_items;
+				ap_destroy_sub_req(subreq);
+
+				/* XXX pkoi pas soptions = local_options ? */
+				soptions |= (local_options & MI_ALLOWFETCH);
+
+				if (local_rss > 0)
+					soptions |= MI_ALLOWRSS;
+			}
 		}
 
 		if (soptions & MI_RECURSIVE) {
 			/* We were asked to recurse, let's dig in */
-			mu_dir			dir_handler;
-			const char		*filename;
-			unsigned short fn_max = MAX_STRING;
-			unsigned short uri_max = MAX_STRING;
+			mu_dir		dir_handler;
+			const char	*filename;
+			unsigned short	fn_max = MAX_STRING;
+			unsigned short	uri_max = MAX_STRING;
+
+			if (!PRETTY_FOLDERS) {
+				subreq = ap_sub_req_lookup_uri(names->uri, r, NULL);
+
+				soptions &= ~(MI_ALLOWFETCH|MI_ALLOWRSS);
+
+				if (subreq != NULL) {
+					local_options = ((mu_config *)ap_get_module_config(subreq->per_dir_config, &musicindex_module))->options;
+					ap_destroy_sub_req(subreq);
+
+					/* XXX pkoi pas soptions = local_options ? */
+					soptions |= (local_options & MI_ALLOWFETCH);
+				}
+			}
 
 			/* We remove the recursive bit at this point.
 			   XXX je comprends pas bien pkoi on | conf->options plutot que soptions */
 			soptions &= ((~(MI_RECURSIVE)) | (conf->options));
 
-			if ( 0 != go_through_directory(r, conf, names) )
-				return head;
+			if (go_through_directory(r, conf, local_options) == FALSE)
+				return;
 
 			/* Try opening the directory */
-			musicindex_opendir(r, &dir_handler, &p, conf, names, soptions);
+			musicindex_opendir(r, &dir_handler, pack, conf, names, soptions);
 			if (NULL == dir_handler.dir)
-				return head;
+				return;
 
 			/* DEV NOTE: keep in mind that uri and fn both points to
 			the end of their "names->*" counterpart */
@@ -398,14 +403,14 @@ mu_ent *make_music_entry(request_rec *r, apr_pool_t *pool,
 					continue;
 				strcpy(fn, filename);
 				strcpy(uri, filename);
-				p = make_music_entry(r, pool, p, conf, names, soptions);
+				make_music_entry(r, pool, pack, conf, names, soptions);
 			}
 			musicindex_closedir(&dir_handler, conf);
-			return p;
+			return;
 		}
 		else if (!(conf->options & MI_STREAM) && !(conf->options & MI_RSS) && !(conf->options & MI_TARBALL)) {
 			if (!(p = NEW_ENT(pool)))
-				return head;
+				return;
 			p->filetype = -1;
 		}
 	}
@@ -413,28 +418,36 @@ mu_ent *make_music_entry(request_rec *r, apr_pool_t *pool,
 		FILE *in = fopen(names->filename, "r");
 
 		if (in == NULL)
-			return head;
+			return;
 
 		/* This is the end of adventures for the array of function pointers.
 		   We call them sequentially, until a function has recognised
 		   it (p!=head), or until we reach the end of the array. In either
 		   case, the file will be closed (this is the magic loop) */
-		for (i=0; (p == head) && make_entry[i]; i++)
-			p = make_entry[i](r, pool, in, head, conf, names);
+		for (i=0; (!p) && make_entry[i]; i++)
+			p = make_entry[i](r, pool, in, conf, names);
 
 	}
 
 	/* The file was not recognised, don't go any further */
-	if (p == head) {
-		return head;
-	}
+	if (!p)
+		return;
 
 	/* OK, now we have an entry, fill in the missing bits */
 	
-	p->next = head; /* Add it to the linked list */
+	p->next = pack->head; /* Add it to the linked list */
 	
 	p->uri = ap_pstrdup(pool, names->uri); /* The absolute URI will be useful at some point */
-	p->file = p->uri; /* Add the relative filename */
+	p->filename = p->file = p->uri; /* Add the relative filename */
+	
+	/* We remove the path prefix if any. We have to treat files differently
+	 * than directories because we can access a given file from various
+	 * parsed_uri.path whereas directories will always be shown from their
+	 * parent directory (XXX assert?) */
+	if ((p->filetype >= 0) && strrchr(p->filename, '/'))
+		p->filename = (strrchr(p->filename, '/') + 1);
+	else if ((p->filetype < 0))
+		p->filename += strlen(r->parsed_uri.path);
 
 	/* convert options into flags - We overwrite whatever was in cache,
 	   as it is an abuse to use these EF_ALLOW flags anyway: nothing like that
@@ -450,46 +463,48 @@ mu_ent *make_music_entry(request_rec *r, apr_pool_t *pool,
 		p->flags |= EF_ALLOWRSS;
 	if ((soptions & MI_CUSTOM) == 0)
 		p->file += strlen(r->parsed_uri.path);  /* remove the path before the filename except for custom playlists */
-	if (conf->options & MI_TARBALL)
+	if (conf->options & MI_TARBALL)	/* if the user asked for a tarball, we're going to need the absolute path to access the file */
 		p->filename = ap_pstrdup(pool, names->filename);
 
-	/* Entry is a directory, don't go further */
-	if (p->filetype < 0)
-		return p;
-
+	/* In the case of directories, we want to have a pretty printed version
+	 * of the directory name, for display and sorting reasons (trailing '/'
+	 * in the real uri would screw alphabetical sorting). Doing the work here
+	 * also saves additional work in html.c. Also on can now assert that
+	 * p->title always exist (sort.c) */
 	if (!(p->title)) {
 #ifdef NO_TITLE_STRIP
-		p->title = p->file;
+		p->title = p->filename;
 #else
 		/* Copy the name removing file extension (if any) and changing '_' to ' ' */
-		char *const stitle = ap_pstrndup(pool, p->file,
-					strrchr(p->file, '.') ?
-					(strrchr(p->file, '.') - p->file) : /* <=> strlen(p->file) - strlen(strrchr(p->file, '.')) */
-					strlen(p->file));
+		char *const stitle = ap_pstrndup(pool, p->filename,
+					((p->filetype >= 0) && strrchr(p->filename, '.')) ?	/* only strip ext if !dir */
+					(strrchr(p->filename, '.') - p->filename) : /* <=> strlen(p->file) - strlen(strrchr(p->file, '.')) */
+					strlen(p->filename));
 		
 		for (i=0; stitle[i]; i++)
 			if (stitle[i] == '_')
 				stitle[i] = ' ';
+		
+		/* Remove trailing '/' (directories) */
+		if (stitle[--i] == '/')
+			stitle[i] = '\0';
 			
 		p->title = stitle; 
 #endif	/* NO_TITLE_STRIP */
-		/* We remove the path prefix if any. */
-		if (strrchr(p->title, '/')) {
-			p->title = strrchr(p->title, '/');
-			p->title++;
-		}
+	}
 
+	/* Entry is a directory, don't go further */
+	if (p->filetype < 0) {
+		pack->dirnb++;
+		pack->head = p;
+		return;
 	}
 
 	/* We put that here so that we do not create cache files for unhandled file types,
 	 * and it's faster to retrieve p->title than to process it again. */
-	if ((p->flags & EF_INCACHE) == 0) {
-		if (conf->cache != NULL) {
-			if (conf->cache->write != NULL)
-				conf->cache->write(r, p, conf, names);
-		}
-	}
-
+	if (conf->cache && (!(p->flags & EF_INCACHE)) && conf->cache->write)
+		conf->cache->write(r, p, conf, names);
+	
 	/* Here we go if we have an ongoing search request, and the file we're
 	 * looking at wasn't found in the cache. We then fall back to a basic strcasestr()
 	 * check, returning only the entries matching the search criteria */
@@ -498,38 +513,45 @@ mu_ent *make_music_entry(request_rec *r, apr_pool_t *pool,
 		if ((ap_strcasestr(file, conf->search))				||
 			(p->artist && (ap_strcasestr(p->artist, conf->search)))	||
 			(p->album && (ap_strcasestr(p->album, conf->search)))	||
-			(p->title && (ap_strcasestr(p->title, conf->search)))		);
+			(p->title && (ap_strcasestr(p->title, conf->search)))	);
 				/* do nothing if there's a match */
 		else	
-			return head;	/* skip if there isn't */
+			return;	/* skip if there isn't */
 	}
 
-	return p;
+	pack->filenb++;
+	pack->head = p;
+	return;
 }
 
 /**
  * Creates a link list based on the custom list.
  *
  * This function generates a list containing all the files referenced in the
- * conf->custom_list string. It handles parameters in a way so that it can
+ * @a conf->custom_list string. It handles parameters in a way so that it can
  * handle indifferently strings for cookies or on-the-fly playlists.
  *
+ * @warning This function uses a subrequest for each file. This has a significant
+ *	performance impact, but OTOH this function isn't really in the 'performance
+ *	path' of the module anyway.
+ *
  * @param r Apache request_rec struct
+ * @param pack A pack to which the generated list will be hooked
  * @param conf MusicIndex configuration paramaters struct
  *
  * @return a list of mu_ent on success, NULL otherwise.
  */
-mu_ent *build_custom_list(request_rec *r, const mu_config *const conf)
+void build_custom_list(request_rec *r, mu_pack *const pack, const mu_config *const conf)
 {
 	request_rec	*subreq;
 	mu_ent_names	names;
 	const char	*args;
 	mu_ent		*mobile_ent = NULL, *custom = NULL, *result = NULL;
-	char		*p, *decodeduri;
+	char		*p, *decodeduri = NULL;
 	short		direct = 0;
 
 	if (conf->custom_list == NULL)
-		return NULL;
+		return;
 
 	args = conf->custom_list;
 
@@ -549,9 +571,9 @@ mu_ent *build_custom_list(request_rec *r, const mu_config *const conf)
 		else if (direct)
 			continue;
 
-		decodeduri = (char *)ap_palloc(r->pool, 1 + ap_base64decode_len(p));
+		decodeduri = (char *)realloc(decodeduri, 1 + ap_base64decode_len(p));
 		if (!decodeduri)
-			return NULL;
+			return;
 		
 		ap_base64decode(decodeduri, p);
 		/* decodeduri is unescaped. We need to escape it before
@@ -572,16 +594,19 @@ mu_ent *build_custom_list(request_rec *r, const mu_config *const conf)
 		ap_destroy_sub_req(subreq);
 		
 		/* adds entries bottom-up */
-		custom = make_music_entry(r, r->pool, custom, conf, &names,
+		make_music_entry(r, r->pool, pack, conf, &names,
 			(MI_CUSTOM | (local_options & MI_ALLOWFETCH)));
 	}
+	free(decodeduri);
 
-	/* reorder the list top-down */
+	/* reorder the list top-down. Deconstify because we do something nasty */
+	custom = (mu_ent *)pack->head;
 	for (result = NULL; custom; custom = mobile_ent) {
-		mobile_ent = custom->next;
+		mobile_ent = (mu_ent *)custom->next;
 		custom->next = result;
 		result = custom;
 	}
+	pack->head = result;
 
-	return result;
+	return;
 }
